package no.uit.mar157.datagenerator;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Random;

/**
 * Generate random data. Contains two parts:
 * 1. makeFastaFile will generate a fasta file with a specified number of random
 * sequences of length between MIN_SEQUENCE_LENGTH and MAX_SEQUENCE_LENGTH.
 * 2. makeBlastFile will generate a simulated blast tabular output with the
 * specified number of rows.
 *
 * Author: Martin Ernstsen
 * Email: martin.ernstsen@gmail.com
 * Date: 7/22/13
 * Time: 2:34 PM
 */
public class DataGenerator {
    private Random rnd;
    private static char[] ACID = "ATCG".toCharArray();
    private static char[] alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789".toCharArray();
    private static int MIN_SEQUENCE_LENGTH = 100;
    private static int MAX_SEQUENCE_LENGTH = 5000;
    private File fastaFile;
    private File blastFile;

    /**
     * Make a new data generator
     * @param fastaFile the fasta output file
     * @param blastFile the simulated blast output file
     */
    public DataGenerator(File fastaFile, File blastFile) {
        this.rnd = new Random(1);
        this.fastaFile = fastaFile;
        this.blastFile = blastFile;
    }

    /**
     * Make a simulated blast output file.
     *
     * @param nRows the number of rows in the output.
     * @throws IOException
     */
    public void makeBlastFile(long nRows) throws IOException {

        try (BufferedWriter bw = new BufferedWriter(new FileWriter(blastFile))) {
            for (int i = 0; i < nRows; i++) {
                bw.write(makeBlastRow());
            }
        }
    }

    private String makeBlastRow() throws IOException {
        StringBuilder row = new StringBuilder();
        row.append(generateID() + "\t")                 //query
        .append(generateID() + "\t")                    //subject
        .append(this.rnd.nextFloat() * 100.0)           //%id
        .append("\t")
        .append(this.rnd.nextInt(MAX_SEQUENCE_LENGTH))  //alignment length
        .append("\t")
        .append(this.rnd.nextInt(MAX_SEQUENCE_LENGTH))  //mismatches
        .append("\t")
        .append(this.rnd.nextInt(MAX_SEQUENCE_LENGTH))  //gap openings
        .append("\t")
        .append(this.rnd.nextInt(MAX_SEQUENCE_LENGTH))  //query start
        .append("\t")
        .append(this.rnd.nextInt(MAX_SEQUENCE_LENGTH))  //query end
        .append("\t")
        .append(this.rnd.nextInt(MAX_SEQUENCE_LENGTH))  //subject start
        .append("\t")
        .append(this.rnd.nextInt(MAX_SEQUENCE_LENGTH))  //subject end
        .append("\t")
        .append(this.rnd.nextFloat())                   //evalue
        .append("\t")
        .append(this.rnd.nextFloat() * 100.0)           //bit score
        .append("\n");

        return row.toString();
    }

    /**
     * Make a fasta file.
     *
     * @param nSequences the number of sequences in the file
     * @throws IOException
     */
    public void makeFastaFile(int nSequences) throws IOException {
         try (BufferedWriter bw = new BufferedWriter(new FileWriter(fastaFile))) {
             for (int i = 0; i < nSequences; i++) {
                 String header = ">" + generateID() + "\n";
                 int sequenceLength = rnd.nextInt(MAX_SEQUENCE_LENGTH - MIN_SEQUENCE_LENGTH) +
                         MIN_SEQUENCE_LENGTH;
                 String sequence = generateSequence(sequenceLength);
                 bw.write(header);
                 bw.write(sequence);
             }
         }
    }

    private String generateSequence(int length) throws IOException {
        StringBuilder sequence = new StringBuilder();
        int remaining = length;
        while (remaining > 60) {
            sequence.append(generateSequenceLine(60));
            remaining -= 60;
        }
        if (remaining > 0) {
            sequence.append(generateSequenceLine(remaining));
        }
        return sequence.toString();
    }

    private String generateSequenceLine(int length) throws IOException {
        StringBuilder line = new StringBuilder();
        for (int i = 0; i < length; i++) {
            line.append(ACID[this.rnd.nextInt(4)]);
        }
        line.append("\n");
        return line.toString();
    }

    /**
     * Make ID's for the fasta and blast files.
     *
     * @return The ID as a string
     */
    private String generateID() {
        StringBuilder id = new StringBuilder();
        for (int i = 0; i < 15; i++) {
            id.append(alphabet[this.rnd.nextInt(alphabet.length)]);
        }
        return id.toString();
    }

//    public static void main(String[] args) {
//        DataGenerator dg = new DataGenerator(new File("abc.fas"), new File("blastres"));
//        try {
////            dg.makeFastaFile(40000);
//            dg.makeBlastFile(1000000);
//        } catch (IOException e) {
//            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
//        }
//    }
}
