package no.uit.mar157.master;

import com.google.gson.Gson;
import no.uit.mar157.common.ClientMasterMessage;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.hbase.HBaseConfiguration;
import org.apache.hadoop.hbase.HColumnDescriptor;
import org.apache.hadoop.hbase.HTableDescriptor;
import org.apache.hadoop.hbase.client.HBaseAdmin;
import org.apache.hadoop.hbase.client.HTable;
import org.apache.hadoop.hbase.client.Put;
import org.apache.hadoop.hbase.io.hfile.Compression;
import org.apache.hadoop.hbase.util.Bytes;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.zeromq.ZMQ;

import java.io.IOException;
import java.sql.*;
import java.util.ArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;


public class Master {
    public static final int MAX_VERSIONS = 50;
    private static Logger logger = LogManager.getLogger(Master.class.getName());
    private static final String ADDR_CONTROLLER = "tcp://127.0.0.1:20999";
    private static final String ADDR_DESC = "tcp://*:20002";
    private static final String ADDR_CLIENT = "tcp://127.0.0.1:20003";
    private static final boolean DELETE_OLD_TABLE = true;

    private ZMQ.Context context;
    private ZMQ.Socket controller;
    private ZMQ.Socket descriptionServer;
    private ZMQ.Socket client;

    protected static final String TABLE_NAME = "mar157_mario";
    protected static final String COL_INPUT = "in";
    protected static final String CF_INPUT = "in";
    protected static final String CF_OUTPUT = "out";
    private Configuration config;
    protected MasterState state;
    protected Object stateNotify;
    private VersionManager versionManager;

    Long startTime;
    Long doneTime;

    protected enum MasterState {READY, RUNNING, PAUSED, DONE}
    public enum MasterCommand {START, PAUSE, STOP, RESET}

    public Master() {
        logger.info("Starting Master...");
        try {
            setupTable(DELETE_OLD_TABLE);
//            makeDummyData();
            versionManager = new VersionManager();
            versionManager.init(DELETE_OLD_TABLE);
        } catch (IOException e) {
            e.printStackTrace();
            logger.error("Error accessing HBase. Shutting down. {}", e.getMessage());
            System.exit(1);
        } catch (SQLException e) {
            logger.error("Error initializing DB connection. Shutting down. {}", e.getMessage());
            System.exit(1);
        }
        setupCommunications();
        state = MasterState.READY;
        stateNotify = new Object();
        startTime = 0L;
        doneTime = 0L;
    }

    private void setupCommunications() {
        context = ZMQ.context(1);
        controller = context.socket(ZMQ.SUB);
        controller.connect(ADDR_CONTROLLER);
        controller.subscribe("KILL".getBytes());
        descriptionServer = context.socket(ZMQ.REP);
        descriptionServer.bind(ADDR_DESC);
        client = context.socket(ZMQ.PULL);
        client.connect(ADDR_CLIENT);
    }

    public ArrayList<Long> getSequenceNumbers() {
        return versionManager.getCurrentVersions();
    }

    private void makeDummyData() throws IOException {
        HTable table = new HTable(config, TABLE_NAME);
        table.setAutoFlush(false);
        for (int i = 0; i < 10; i++) {
            Put put = new Put(Bytes.toBytes("rowkey" + String.valueOf(i)));
            put.add(Bytes.toBytes(CF_INPUT), Bytes.toBytes(COL_INPUT), 0L, Bytes.toBytes("data" + String.valueOf(i)));
            table.put(put);
        }
        table.flushCommits();
        table.close();
    }

    private void resetTable() throws IOException {
        HBaseAdmin admin = new HBaseAdmin(config);
        admin.disableTable(TABLE_NAME);
        admin.deleteColumn(TABLE_NAME, CF_OUTPUT);

        HColumnDescriptor outpuColumnFamilyDescriptor = new HColumnDescriptor(Bytes.toBytes(CF_OUTPUT));
        outpuColumnFamilyDescriptor.setCompressionType(Compression.Algorithm.SNAPPY);
//        outpuColumnFamilyDescriptor.setMaxVersions(MAX_VERSIONS);
        admin.addColumn(TABLE_NAME, outpuColumnFamilyDescriptor);
        admin.enableTable(TABLE_NAME);
        admin.close();
    }

    private void setupTable(boolean deleteOldTable) throws IOException {
        config = HBaseConfiguration.create();
        if (deleteOldTable) {
            HBaseAdmin admin = new HBaseAdmin(config);
            if (admin.tableExists(TABLE_NAME)) {
                admin.disableTable(TABLE_NAME);
                admin.deleteTable(TABLE_NAME);
            }
            HColumnDescriptor inputColumnFamilyDescriptor = new HColumnDescriptor(Bytes.toBytes(CF_INPUT));
            inputColumnFamilyDescriptor.setCompressionType(Compression.Algorithm.SNAPPY);
            HColumnDescriptor outpuColumnFamilyDescriptor = new HColumnDescriptor(Bytes.toBytes(CF_OUTPUT));
            outpuColumnFamilyDescriptor.setCompressionType(Compression.Algorithm.SNAPPY);
//            outpuColumnFamilyDescriptor.setMaxVersions(MAX_VERSIONS);

            HTableDescriptor tableDescriptor = new HTableDescriptor(TABLE_NAME);
            tableDescriptor.setDeferredLogFlush(true);
            tableDescriptor.addFamily(inputColumnFamilyDescriptor);
            tableDescriptor.addFamily(outpuColumnFamilyDescriptor);
            admin.createTable(tableDescriptor);
            admin.close();
        }
    }

    private void run() throws InterruptedException {
        logger.info("Master running");

        ExecutorService threadExecutor = Executors.newSingleThreadExecutor();
        Ventilator vent = new Ventilator(this, context, config);
        Future ventFuture = threadExecutor.submit(vent);

        Gson gson = new Gson();
        ZMQ.Poller poller = new ZMQ.Poller(3);
        poller.register(controller, ZMQ.Poller.POLLIN);
        poller.register(descriptionServer, ZMQ.Poller.POLLIN);
        poller.register(client, ZMQ.Poller.POLLIN);
        logger.debug("Entering polling loop");
        while (!Thread.currentThread().isInterrupted()) {
            poller.poll();
            if (poller.pollin(0)) {
                byte[] rawMessage = controller.recv(0);
                String messageString = new String(rawMessage);
                logger.debug("Received control message: " + messageString);
                break;
            }
            if (poller.pollin(1)) {
                //request for pipeline description
                descriptionServer.recv();
                logger.debug("Received request for pipelinedescription");
                descriptionServer.send(gson.toJson(versionManager.getPipelineDescription()));
            }
            if (poller.pollin(2)) {
                //message from client
                byte[] rawMessage1 = client.recv(0);
                String messageString = new String(rawMessage1);

                ClientMasterMessage clientMasterMessage = gson.fromJson(messageString, ClientMasterMessage.class);
                if (clientMasterMessage.getPipelineDescription() != null) {
                    try {
                        versionManager.update(clientMasterMessage.getPipelineDescription());
                    } catch (SQLException e) {
                        e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                    }
                }
                if (state == MasterState.READY && clientMasterMessage.getCommand() == MasterCommand.START) {
                    if (versionManager.getPipelineDescription() != null) {
                        logger.info("Start command received, state is READY");
                        setState(MasterState.RUNNING);
                    } else {
                        logger.error("Trying to start without pipeline description");
                    }
                }
                else if (state == MasterState.PAUSED && clientMasterMessage.getCommand() == MasterCommand.START) {
                    logger.info("Start (resume) command received, state is PAUSED");
                    setState(MasterState.RUNNING);
                }
                else if (state == MasterState.DONE && clientMasterMessage.getCommand() == MasterCommand.START) {
                    logger.info("Start command received, state is DONE");
                    setState(MasterState.RUNNING);
                }
                else if (state == MasterState.RUNNING && clientMasterMessage.getCommand() == MasterCommand.PAUSE) {
                    logger.info("Pause command received");
                    setState(MasterState.PAUSED);
                }
                else if ((state == MasterState.DONE || state == MasterState.PAUSED) &&
                        clientMasterMessage.getCommand() == MasterCommand.RESET) {
                    logger.info("Reset command received");
                    try {
                        resetTable();
                        setState(MasterState.READY);
                    } catch (IOException e) {
                        logger.error("Unable to reset HBase table: {}", e.getMessage());
                    }
                }
            }
        }
        // Clean up and shut down
        boolean result = ventFuture.cancel(true);
        logger.debug("result of ventilator shutdown: {}", result);
        versionManager.close();
        threadExecutor.shutdown();
        controller.close();
        descriptionServer.close();
        client.close();
        context.term();
        logger.info("Shutting down master");
    }

    protected MasterState getState() {
        return state;
    }

    protected synchronized void setState (MasterState newState) {
        logger.debug("Changing state from {} to {}", state, newState);
        if (newState == MasterState.RUNNING) {
            synchronized (stateNotify) {
                if (state == MasterState.READY || state == MasterState.DONE) {
                    startTime = System.currentTimeMillis();
                }
                state = newState;
                stateNotify.notify();
            }
        } else {
            if (state == MasterState.RUNNING && newState == MasterState.DONE) {
                doneTime = System.currentTimeMillis();
                logger.info("Time elapsed from START until DONE: {}ms / {}s", doneTime - startTime, (doneTime - startTime) / 1000);
            }
            state = newState;
        }
    }

    public static void main(String[] args) {
        try {
            Master m = new Master();
            m.run();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
    }
}
