package no.uit.mar157.master;

import com.google.gson.Gson;
import no.uit.mar157.common.TaskMessage;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.hbase.client.HTable;
import org.apache.hadoop.hbase.client.Result;
import org.apache.hadoop.hbase.client.ResultScanner;
import org.apache.hadoop.hbase.client.Scan;
import org.apache.hadoop.hbase.util.Bytes;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.zeromq.ZMQ;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Random;


/**
 * The Ventilator is the thread in the Master process, whose main responsibility is to
 * distribute work to the workers, by sending TaskMessages. Ventilator is a ZeroMQ term.
 */
public class Ventilator implements Runnable {
    private static final String ADDR_VENTILATOR = "tcp://*:20000";
    private static final String ADDR_STARTUPSYNC = "tcp://*:20001";
    private static final String ADDR_WORKERACK = "tcp://*:20004";
    private static final int WORKER_COUNT = 8;
    private static final int BATCH_SIZE = WORKER_COUNT * 2;
    private static Logger logger = LogManager.getLogger(Ventilator.class.getName());
    private ZMQ.Context context;
    private ZMQ.Socket startupSync;
    private ZMQ.Socket ventilator;
    private ZMQ.Socket workerAck;

    private Configuration config;
    private Master master;

    private static final int SCAN_CACHING = 500;
    private static boolean USE_RESERVOIR_SAMPLING = false;
    private static int SAMPLE_SIZE = 10000;
    private Random random;
    private ArrayList<String> sample;


    public Ventilator(Master master, ZMQ.Context ctx, Configuration hbaseConf) {
        this.master = master;
        context = ctx;
        config = hbaseConf;
        random = new Random();
    }


    @Override
    public void run() {
        logger.info("Starting ventilator thread...");
        setupSockets();

        for (int i = 0; i < WORKER_COUNT; i++) {
            byte[] req = startupSync.recv(0);
            logger.debug("Syncing worker {}", i);
            logger.debug("Received startup sync message: {}", new String(req));
            logger.debug("Replying to startup sync message");
            startupSync.send("OK".getBytes(), 0);
        }

        try (HTable table = new HTable(config, Master.TABLE_NAME)) {
            while (true) {
                logger.debug("Ventilator waiting");
                checkState();
                logger.debug("Ventilator running");
                ArrayList<Long> cv = master.getSequenceNumbers();
                int remainingAcks = 0;

                Scan scan = new Scan();
                scan.setCaching(SCAN_CACHING);
                scan.addColumn(Bytes.toBytes(Master.CF_INPUT), Bytes.toBytes(Master.COL_INPUT));
                ResultScanner scanner = table.getScanner(scan);

                if (USE_RESERVOIR_SAMPLING) {
                    remainingAcks = ventilateSample(cv, scanner);
                } else {
                    remainingAcks = ventilateAll(cv, scanner);
                }

                // Receive remaining workeracks
                for (int i = 0; i <  remainingAcks; i++) {
                    workerAck.recv();
                }
                master.setState(Master.MasterState.DONE);
            }
        } catch (IOException e) {
            logger.error(e.getMessage());
            e.printStackTrace();
        } catch (InterruptedException e) {
            logger.debug("Ventilator interrupted: {}", e.getMessage());
        } finally {
            logger.debug("closing ventilator sockets");
            closeSockets();
        }
    }

    private int ventilateSample(ArrayList<Long> cv, ResultScanner scanner) throws InterruptedException {
        if (sample == null) {
            sample = new ArrayList<>(SAMPLE_SIZE);
            long index = 0L;
            for (Result result : scanner) {
                if (index < SAMPLE_SIZE) {
                    sample.add(new String(result.getRow()));
                } else {
                    long r = (long)(random.nextDouble() * index);
                    if (r < SAMPLE_SIZE) {
                        sample.set((int)r, new String(result.getRow()));
                    }
                }
                index++;
            }
            if (index < SAMPLE_SIZE) {
                logger.error("Sampling error: dataset smaller than sample size");
                // TODO Error handling
            }
        }
//        logger.debug("sample array: {}", sample);

        Gson gson = new Gson();
        int batchCount = 0;
        for (String key : sample) {
            checkState();
            if (batchCount == BATCH_SIZE) {
                for (int i = 0; i < BATCH_SIZE; i++) {
                    workerAck.recv();
                }
                batchCount = 0;
            }
            TaskMessage msg = new TaskMessage(key, cv);
//            logger.debug("sending key: {}", key);
            boolean ret = ventilator.send(gson.toJson(msg), 0);
            if (ret) {
//                        logger.debug("Ventilator transmitting: {}", gson.toJson(msg));
            } else {
                logger.debug("Ventilator unsuccessful. Key: {}", gson.toJson(msg));
            }
            batchCount++;
        }
        return batchCount;
    }

    private int ventilateAll(ArrayList<Long> cv, ResultScanner scanner) throws InterruptedException {
        Gson gson = new Gson();
        int batchCount = 0;
        for (Result result : scanner) {
            checkState();
            if (batchCount == BATCH_SIZE) {
                for (int i = 0; i < BATCH_SIZE; i++) {
                    workerAck.recv();
                }
                batchCount = 0;
            }
            byte[] key = result.getRow();
            TaskMessage msg = new TaskMessage(new String(key), cv);
            boolean ret = ventilator.send(gson.toJson(msg), 0);
            if (ret) {
//                        logger.debug("Ventilator transmitting: {}", gson.toJson(msg));
            } else {
                logger.debug("Ventilator unsuccessful. Key: {}", gson.toJson(msg));
            }
            batchCount++;
        }
        return batchCount;
    }

    void checkState() throws InterruptedException {
        synchronized (master.stateNotify) {
            Master.MasterState state = master.getState();
            switch (state) {
                case READY:
                    master.stateNotify.wait();
                    break;
                case PAUSED:
                    master.stateNotify.wait();
                    break;
                case DONE:
                    master.stateNotify.wait();
                    break;
                default:
                    break;
            }
        }
    }

    void setupSockets() {
        ventilator = context.socket(ZMQ.PUSH);
        ventilator.bind(ADDR_VENTILATOR);

        startupSync = context.socket(ZMQ.REP);
        startupSync.bind(ADDR_STARTUPSYNC);

        workerAck = context.socket(ZMQ.PULL);
        workerAck.bind(ADDR_WORKERACK);
    }

    private void closeSockets() {
        startupSync.close();
        ventilator.close();
        workerAck.close();
    }
}
