package no.uit.mar157.worker;

import com.google.gson.Gson;
import no.uit.mar157.common.PipelineDescription;
import no.uit.mar157.common.PipelineStage;
import no.uit.mar157.common.TaskMessage;
import org.apache.hadoop.hbase.client.Get;
import org.apache.hadoop.hbase.client.HTable;
import org.apache.hadoop.hbase.client.Put;
import org.apache.hadoop.hbase.client.Result;
import org.apache.hadoop.hbase.util.Bytes;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.zeromq.ZMQ;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.Callable;


/**
 * Taskrunner is a Worker thread. It runs one key through the whole pipeline.
 */
public class TaskRunner extends Thread {
    private static Logger logger = LogManager.getLogger(TaskRunner.class.getName());
    private static final String TMP_BASEDIR = "/tmp/mar157_mario";
    private final String TMP_DIR;
    protected static final String TABLE_NAME = "mar157_mario";
    protected static final String COL_INPUT = "in";
    protected static final String CF_INPUT = "in";
    protected static final String CF_OUTPUT = "out";

    private final String port;
    private PipelineDescription pipelineDescription;
    HTable table;

    private ZMQ.Context context;


    public TaskRunner(String port, HTable table, ZMQ.Context context) {
        this.pipelineDescription = null;
        this.port = port;
        this.table = table;
        this.TMP_DIR = TMP_BASEDIR + port;
        this.context = context;
        makeTmpDir();
    }

    private void makeTmpDir() {
        try {
            File tmpDir = new File(TMP_DIR);
            if (!tmpDir.exists()) {
                if (tmpDir.mkdir()) {
                    logger.info(port + ": Created working directory: {}", TMP_DIR);
                } else {
                    throw new IOException("Unable to create working directory: " + TMP_DIR);
                }
            }
        } catch (IOException e) {
            logger.error("TaskRunner error: {}", e.getMessage());
        }
    }

    @Override
    public void run() {
        logger.info("Starting TaskRunner thread");
        ZMQ.Socket taskReceiver = context.socket(ZMQ.PAIR);
        taskReceiver.connect("inproc://task");
        ZMQ.Socket descriptionReceiver = context.socket(ZMQ.PAIR);
        descriptionReceiver.connect("inproc://description");
        ZMQ.Socket killer = context.socket(ZMQ.PAIR);
        killer.connect("inproc://killer");
        ZMQ.Socket notifications = context.socket(ZMQ.PAIR);
        notifications.connect("inproc://notifications");

        ZMQ.Poller poller = new ZMQ.Poller(3);
        poller.register(killer, ZMQ.Poller.POLLIN);
        poller.register(descriptionReceiver, ZMQ.Poller.POLLIN);
        poller.register(taskReceiver, ZMQ.Poller.POLLIN);

        Gson gson = new Gson();
        while (!Thread.currentThread().isInterrupted()) {
            poller.poll();
            if (poller.pollin(0)) {
                killer.recv();
                break;
            }
            if (poller.pollin(1)) {
                byte[] rawMessage = descriptionReceiver.recv();
                String messageString = new String(rawMessage);
//                logger.debug("TaskRunner Received description from Worker: {}", messageString);
                setPipelineDescription(gson.fromJson(messageString, PipelineDescription.class));
            }
            if (poller.pollin(2)) {
                byte[] rawMessage = taskReceiver.recv();
                String message = new String(rawMessage);
                TaskMessage taskMessage = gson.fromJson(message, TaskMessage.class);
                doTask(taskMessage);
                notifications.send("done", ZMQ.DONTWAIT);
            }
        }
        logger.info("TaskRunner shutting down");
        taskReceiver.close();
        descriptionReceiver.close();
        killer.close();
        notifications.close();
    }

    private void doTask(TaskMessage taskMessage) {
//        logger.debug("TaskRunner doing task: {}", taskMessage.toString());
        try {
            boolean isFirstTimeWritingToDisk = true;
            Get get = new Get(Bytes.toBytes(taskMessage.getId()));
            Result row = table.get(get);
            ArrayList<Long> cv = taskMessage.getCurrentVersions();
            String inputColumn = null;
            String inputCF = null;
            String outputColumn = null;
            File tmpDir = new File(TMP_DIR);
            File inFile = null;
            File outFile = null;

            for (int i = 0; i < pipelineDescription.size(); i++) {
                PipelineStage stage = pipelineDescription.getStage(i);
                if (i == 0) {
                    // First stage of pipeline, so parent column is COL_INPUT, parent version is 0
                    inputColumn = COL_INPUT;
                    inputCF = CF_INPUT;
                    int stageNumber = i + 1;
                    Long parentVersion = 0L;
                    Long version = cv.get(i);
                    outputColumn = "out_" + String.valueOf(stageNumber) + "_" +
                            String.valueOf(parentVersion) + "_" +
                            String.valueOf(version);
                } else {
                    inputColumn = outputColumn; // Input to this stage is output from previous stage
                    inputCF = CF_OUTPUT;
                    int stageNumber = i + 1;
                    Long parentVersion = cv.get(i - 1);
                    Long version = cv.get(i);
                    outputColumn = "out_" + String.valueOf(stageNumber) + "_" +
                            String.valueOf(parentVersion) + "_" +
                            String.valueOf(version);
                }

                if (row.containsColumn(Bytes.toBytes(CF_OUTPUT), Bytes.toBytes(outputColumn))) {
                    //                logger.debug(port + ": Skipping computation of outputcolumn: {}", outputColumn);
                    continue;   // This output has already been computed
                }

                // Must compute output
                // Get input data and write to file
                if (isFirstTimeWritingToDisk) {
                    byte[] inputValue = row.getValue(Bytes.toBytes(inputCF), Bytes.toBytes(inputColumn));
                    inFile = new File(tmpDir, inputColumn);
                    FileOutputStream os = new FileOutputStream(inFile);
                    os.write(inputValue);
                    os.close();
                    outFile = new File(tmpDir, outputColumn);
                    isFirstTimeWritingToDisk = false;
                } else {
                    inFile = outFile;
                    outFile = new File(tmpDir, outputColumn);
                }

                // Run the command, abort on error
                if (runExternalProcess(tmpDir, inFile, outFile, stage) != 0) break;  // Stop execution

                // Put output to HBase
                byte[] data = Files.readAllBytes(outFile.toPath());
                Put put = new Put(Bytes.toBytes(taskMessage.getId()));
                put.add(Bytes.toBytes(CF_OUTPUT), Bytes.toBytes(outputColumn), data);
                table.put(put);
                if (!inFile.delete()) {
                    logger.error(port + ": Unable to delete tmp file: {}", inFile.getName());
                }
            }
            table.flushCommits();
        } catch (IOException e) {
            logger.error("Exception in Taskrunner: {}", e.getMessage());
        } catch (InterruptedException e) {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }
//        logger.debug("TaskRunner completed task");
    }

    private int runExternalProcess(File tmpDir, File inFile, File outFile, PipelineStage stage) throws IOException, InterruptedException {
        // Replace {in} and {out} with filenames, if specified
        String command = stage.getCommand();
        boolean replaceIn = command.contains("{in}");
        boolean replaceOut = command.contains("{out}");
        if (replaceIn && replaceOut) {
            command = command.replaceAll("\\{in\\}", inFile.getAbsolutePath());
            command = command.replaceAll("\\{out\\}", outFile.getAbsolutePath());
        } else if (replaceIn) {
            command = command.replaceAll("\\{in\\}", inFile.getAbsolutePath());

        } else if (replaceOut) {
            command = command.replaceAll("\\{out\\}", outFile.getAbsolutePath());
        }

        // Construct command to run
        List<String> cmd = new ArrayList<>(Arrays.asList("bash", "-c", command));
        cmd.addAll(stage.getParameterList());

        // Get the full command string for the logger
        StringBuilder fullCmdBuilder = new StringBuilder();
        for(String part: cmd) {
            fullCmdBuilder.append(part).append(" ");
        }
//        logger.debug(port + ": Running command: {}", fullCmdBuilder.toString());
        Process p;
        if (replaceIn && replaceOut) {
            p = new ProcessBuilder()
                    .directory(tmpDir)
                    .redirectErrorStream(true)
                    .command(cmd)
                    .start();
        } else if (replaceIn) {
            p = new ProcessBuilder()
                    .directory(tmpDir)
                    .redirectErrorStream(true)
                    .redirectOutput(outFile)
                    .command(cmd)
                    .start();
        } else if (replaceOut) {
            p = new ProcessBuilder()
                    .directory(tmpDir)
                    .redirectErrorStream(true)
                    .redirectInput(inFile)
                    .command(cmd)
                    .start();
        } else {
            p = new ProcessBuilder()
                    .directory(tmpDir)
                    .redirectErrorStream(true)
                    .redirectInput(inFile)
                    .redirectOutput(outFile)
                    .command(cmd)
                    .start();
        }
        int exitValue = p.waitFor();
        if (exitValue != 0) {
//                    status = i+1; // TODO something must be done with stage numbering
//            logger.error(port + ": Error executing stage {}", i);
        }
        return exitValue;
    }

    private void setPipelineDescription(PipelineDescription description) {
        pipelineDescription = description;
    }
}


