package no.uit.mar157.worker;

import com.google.gson.Gson;
import no.uit.mar157.common.PipelineDescription;
import no.uit.mar157.common.TaskMessage;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.hbase.HBaseConfiguration;
import org.apache.hadoop.hbase.client.HTable;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.zeromq.ZMQ;

import java.io.IOException;
import java.util.ArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;


public class Worker {
    private static Logger logger = LogManager.getLogger(Worker.class.getName());
//    private static final String ADDR_CONTROLLER = "tcp://127.0.0.1:20999";
//    private static final String ADDR_STARTUPSYNC = "tcp://127.0.0.1:20001";
//    private static final String ADDR_VENTILATOR = "tcp://127.0.0.1:20000";
//    private static final String ADDR_DESC = "tcp://127.0.0.1: 20002";
//    private static final String ADDR_WORKERACK = "tcp://127.0.0.1:20004";

    private static final String ADDR_CONTROLLER = "tcp://compute-0-0:20999";
    private static final String ADDR_STARTUPSYNC = "tcp://compute-0-0:20001";
    private static final String ADDR_VENTILATOR = "tcp://compute-0-0:20000";
    private static final String ADDR_DESC = "tcp://compute-0-0: 20002";
    private static final String ADDR_WORKERACK = "tcp://compute-0-0:20004";

    protected static final String TABLE_NAME = "mar157_mario";

    private enum TaskRunnerState {AVAILABLE, BUSY}

    private static String port;
    private PipelineDescription pipelineDescription;
    private int progressCounter;

    private ZMQ.Context context;
    private ZMQ.Socket controller;
    private ZMQ.Socket startupSync;
    private ZMQ.Socket ventilator;
    private ZMQ.Socket descriptionRequester;
    private ZMQ.Socket workerAck;

    private ZMQ.Socket taskSender;
    private ZMQ.Socket descriptionSender;
    private ZMQ.Socket killer;
    private ZMQ.Socket notifications;

    public Worker(String ownPort) {
        port = ownPort;
        logger.info(port + ": Starting Worker...");
        context = ZMQ.context(1);
        controller = context.socket(ZMQ.SUB);
        controller.setLinger(1);
        controller.connect(ADDR_CONTROLLER);
        controller.subscribe("KILL".getBytes());
        startupSync = context.socket(ZMQ.REQ);
        startupSync.setLinger(1);
        startupSync.connect(ADDR_STARTUPSYNC);
        ventilator = context.socket(ZMQ.PULL);
        ventilator.setLinger(1);
        ventilator.connect(ADDR_VENTILATOR);
        descriptionRequester = context.socket(ZMQ.REQ);
        descriptionRequester.setLinger(1);
        descriptionRequester.connect(ADDR_DESC);
        workerAck = context.socket(ZMQ.PUSH);
        workerAck.setLinger(1);
        workerAck.connect(ADDR_WORKERACK);

        taskSender = context.socket(ZMQ.PAIR);
        taskSender.bind("inproc://task");
        descriptionSender = context.socket(ZMQ.PAIR);
        descriptionSender.bind("inproc://description");
        killer = context.socket(ZMQ.PAIR);
        killer.bind("inproc://killer");
        notifications = context.socket(ZMQ.PAIR);
        notifications.bind("inproc://notifications");

        pipelineDescription = null;
        progressCounter = 0;
    }

    private boolean hasLatestPipelineDescription(ArrayList<Long> latestVersion) {
        boolean retVal;
        if (pipelineDescription == null) {
            retVal = false; // this happens on initialization
        } else {
            ArrayList<Long> workerVersion = pipelineDescription.getSequenceNumbers();
            retVal = workerVersion.equals(latestVersion);
        }
        return retVal;
    }

    private byte[] requestNewPipelineDescription() {
        Gson gson = new Gson();
        logger.debug(port + ": Requesting new description");
        descriptionRequester.send("");
        byte[] rawMessage = descriptionRequester.recv();
        String messageString = new String(rawMessage);
        logger.debug(port + ": Received description: {}", messageString);
        pipelineDescription = gson.fromJson(messageString, PipelineDescription.class);
        return rawMessage;
    }

    private void run() {
        logger.info(port + ": Worker running");

        Configuration config = HBaseConfiguration.create();
        config.set("hbase.zookeeper.quorum", "compute-0-0");
        HTable table = null;
        try {
            table = new HTable(config, TABLE_NAME);
        } catch (IOException e) {
            logger.error(port + ": Error. Unable to create HTable. Shutting down. {}", e.getMessage());
            System.exit(1);
        }
        table.setAutoFlush(false);

        // Tell master that we are ready TODO, this should be improved
        logger.debug("Sending startup sync message");
        startupSync.send("READY".getBytes(), 0);
        byte[] okReply = startupSync.recv(0);
        logger.debug("Received {} from Master", new String(okReply));

        TaskRunnerState state = TaskRunnerState.AVAILABLE;
        TaskRunner taskRunner = new TaskRunner(port, table, context);
        taskRunner.start();

        ZMQ.Poller poller = new ZMQ.Poller(3);
        poller.register(controller, ZMQ.Poller.POLLIN);
        poller.register(ventilator, ZMQ.Poller.POLLIN);
        poller.register(notifications, ZMQ.Poller.POLLIN);

        Gson gson = new Gson();
        logger.debug("Worker entering polling loop");
        while (!Thread.currentThread().isInterrupted()) {
            poller.poll();
            if (poller.pollin(0)) {
                killer.send("kill", ZMQ.DONTWAIT);
                byte[] rawMessage = controller.recv(ZMQ.DONTWAIT);
                String message = new String(rawMessage);
                logger.debug(port + ": Received message: " + message);
                try {
                    // Give some time for the TaskRunner thread to shut down
                    Thread.sleep(1000);
                } catch (InterruptedException e) {
                    logger.error("Error shutting down: {}", e.getMessage());
                }
                break;
            }
            if (poller.pollin(1)) {
                if (state == TaskRunnerState.AVAILABLE) {
                    byte[] rawMessage = ventilator.recv();
                    String message = new String(rawMessage);
                    TaskMessage taskMessage = gson.fromJson(message, TaskMessage.class);

                    // check pipeline version. If old, request new
                    if (!hasLatestPipelineDescription(taskMessage.getCurrentVersions())) {
                        byte[] desc = requestNewPipelineDescription();
                        logger.debug("sending new description to taskrunner");
                        descriptionSender.send(desc, ZMQ.DONTWAIT);
                    }
                    state = TaskRunnerState.BUSY;
                    taskSender.send(message, ZMQ.DONTWAIT);

                    progressCounter++;
                    if (progressCounter == 999) {
                        logger.info("Processed 1000 rows...");
                        progressCounter = 0;
                    }
                }
            }
            if (poller.pollin(2)) {
                workerAck.send("", ZMQ.DONTWAIT);
                state = TaskRunnerState.AVAILABLE;
            }
        }

        // Clean up and shut down
        try {
            table.close();
        } catch (IOException e) {
            logger.error(port + ": Error closing HTable. {}", e.getMessage());
        }
        controller.close();
        startupSync.close();
        ventilator.close();
        descriptionRequester.close();
        workerAck.close();

        taskSender.close();
        descriptionSender.close();
        killer.close();
        notifications.close();

        context.term();
        logger.debug("at least context is shut down...");
        logger.info(port + ": Shutting down worker");
    }

    public static void main(String[] args) {
        if (args.length != 1) {
            System.err.println(
                    "Error: wrong number of arguments. Usage\n" +
                            "Java ...worker... port number");
            System.exit(1);
        }
        Worker w = new Worker(args[0]);
        w.run();
    }
}
